#include <windows.h>
#include <imagehlp.h>
#include <alloc.h>
#include <dos.h>
#include <stdio.h>
#include <malloc.h>
#include <imagehlp.h>
#include <tlhelp32.h>
#include <vcl\system.hpp>


// WATCH2
// by Lucian Wischik
// This code is based heavily upon the MEMWATCH program by Terry Richards,
// and some sample code in the Microsoft Systems Journal vol12 num5
//
// This code monitors the following memory allocations:
//  - calls you make to new, new[], delete, delete[]
//  - The GetMem, ReallocMem, and FreeMem standard procedures.
//  - The New and Dispose standard procedures.
//  - Allocation and deallocation of objects through constructors and destructors,
//    including all VCL objects
//  - Allocation and deallocation of dynamic strings.
//  - Also it checks that you haven't overwritten or underwritten your memory blocks
//  - It does NOT monitor Window system calls, and does NOT check that you have
//    freed all Windows objects that you create.
//
// To use, simply add the file watch2.cpp to your project and rebuild. It
// does not work at all well with dynamic linking. I recommend that you turn
// it off. If you're using it in a package, you must set WatchVcl=false;
// There are two functions you might find useful (but will have to prototype yourself)
//   bool __fastcall CheckSafetyAreas(void *ptr); // check safety areas not overwritten
//   AnsiString __fastcall dcallstack(); // retrieve the current callstack
//=============================================================================

const char*LogName="*c:\\otabrowsememlog.txt"; // do we output a log to a file?
const bool ReportNulls=true; // do we report attempts to delete NULLs?
const bool NagScreen=true;   // a dialog box that pops up if there was any leakage
const bool WatchVcl=false;   // keep track of vcl stuff? should be false for a package
const int  NumEnts = 5000;   // biggest number of memory blocks that can be allocated
const char SafetyChar='!';   // Safety area is to check if you had memory oversteps
const int  HiSafetySize=20;  //
const int  LoSafetySize=0;   // Apparently it has to be 0 if you use dynamic rtl
const int  MaxStackDepth=30; // How much call-stack do we record?
const int  MaxModules=50;    // How many modules will we be willing to inspect?

//=============================================================================
// This is how it intercepts memory allocations
// - It provides its own version of operator new. This means that any time you
//   call a regular C++ new operator, it in fact calls the version in this code.
// - It installs a startup procedure with #pragma startup MyStartupProc 0.
//   This startup proc gets called right at the very start. In MyStartupProc,
//   it calls SetMemoryManager(...) This instructs the VCL to use the memory
//   allocation routines provided by Watch2.
//
// This is how it keeps track of leaked memory
// - There is a large array of size NumEnts. Whenever new is called, a record is
//   made in this array. Whenever delete is called, it removes the record from the array
// - The large array records only the address of each memory block allocated, and its size.
// - We actually allocate more memory than is strictly needed. We allocate
//   LoSafetySize bytes beneath it, and HiSafetySize bytes above it, and we fill them with
//   SafetyChar. This is so that, when it comes time to delete the block, we can verify
//   that it hasn't accidentally been overwritten
// - In fact, we actually make each memory block even bigger. Extra space is reserved
//   at the top of the high safety block and this is filled with a record of the call-stack
//   at the time the block was allocated. That is, a TCallstackHeader structure followed
//   by some number of TLineOfCallstack structures.
// - An exit procedure is installed with #pragma exit MyExitProc 0. This gets called
//   right at the very end. At this stage we go through the large array and pick out
//   any blocks that have not yet been deleted. The user is informed.
//
// This is a note on how you should use the results
// - Watch displays the callstack how it was when a memory block was allocated.
//   The top entries are the most recent.
// - Allocating VCL objects is kind of strange. Although you think you called
//   new TButton(..) from within your Form constructor, or from within
//   one of your methods, in fact the VCL call to memory allocation probably happened
//   at quite a different time. This can make it very difficult to track down
//   the exact reason for the call. I have no idea why this is so.
// - There's an extra function you can use. bool CheckSafetyAreas(void *)
//   If you want to use it in your own code you'll have to prototype it.
//   It ensures that the safety areas above and below a block of memory have
//   not been corrupted.
// - If you do use Watch, then your programs will run slower
//   and will take a bit longer to start up. This is unavoidable.
//
// These are particular implementation notes
// - The RTL itself has some memory leaks, of about 1912 bytes. It would be a pain
//   if these were always reported. Hence, Watch does not take them seriously.
//   It does this my installing #pragma startup MyStartTakingSeriouslyProc 32.
//   If any memory blocks were allocated before startup-level-32, Watch does not
//   complain if they are not deleted. But if anything is allocated after startup-level-32
//   then Watch does complain. I found the number '32' experimentally.
// - The Windows IMAGEHLP routines are used to obtain the stack trace.
//






typedef struct {char c[MAX_PATH]; void *pc; void *frame;} TLineOfCallstack;
typedef struct {int Depth; bool Serious; TLineOfCallstack l[MaxStackDepth];} TCallstack;
typedef struct {void *Address; size_t Size;} TMemoryBlock;
typedef struct {char imageName[MAX_PATH]; char moduleName[MAX_PATH]; DWORD baseAddress; DWORD size;} TModuleEntry;


TMemoryBlock Mems[NumEnts]; // The database of memory entries
bool MemsFull=false;        // Have we ever run out of space in the database?
bool DontTakeSeriously=true;// the rtl doesn't free all of its memory: so, we don't take the first calls seriously
bool InternalCall=false;    // to prevent recursion
int  MaxUsed=0,NumNew=0,NumDeletes=0,NumNullDeletes=0,NewCalled=0; // statistics

TModuleEntry Mods[MaxModules]; // The database of modules currently loaded
int NumMods=0;                 // number of modules in the database

bool db_isinited=false; // have we attempted at least once to start logging?
FILE *db_f;             // log file handle

bool sym_isinited=false;// have we tried at least once to init?
bool sym_issucc=false;  // was it succesfull?
bool sym_usemods=false; // will we also be able to enumerate modules?
bool sym_usesyms=false; // will we also be able to use symbols?
HINSTANCE sym_hPsapi = NULL;
HINSTANCE sym_hImagehlpDll=NULL;
HINSTANCE sym_hToolHelp = NULL;










typedef BOOL (__stdcall *SYMCLEANUPPROC)( IN HANDLE hProcess );
typedef PVOID (__stdcall *SYMFUNCTIONTABLEACCESSPROC)( HANDLE hProcess, DWORD AddrBase );
typedef BOOL (__stdcall *SYMGETLINEFROMADDRPROC)( IN HANDLE hProcess, IN DWORD dwAddr,	OUT PDWORD pdwDisplacement, OUT PIMAGEHLP_LINE Line );
typedef DWORD (__stdcall *SYMGETMODULEBASEPROC)( IN HANDLE hProcess, IN DWORD dwAddr );
typedef BOOL (__stdcall *SYMGETMODULEINFOPROC)( IN HANDLE hProcess, IN DWORD dwAddr, OUT PIMAGEHLP_MODULE ModuleInfo );
typedef DWORD (__stdcall *SYMGETOPTIONSPROC)( VOID );
typedef BOOL (__stdcall *SYMGETSYMFROMADDRPROC)(IN HANDLE hProcess, IN DWORD dwAddr,	OUT PDWORD pdwDisplacement, OUT PIMAGEHLP_SYMBOL Symbol );
typedef BOOL (__stdcall *SYMINITIALIZEPROC)( IN HANDLE hProcess, IN PSTR UserSearchPath, IN BOOL fInvadeProcess );
typedef DWORD (__stdcall *SYMLOADMODULEPROC)(IN HANDLE hProcess, IN HANDLE hFile,IN PSTR ImageName, IN PSTR ModuleName, IN DWORD BaseOfDll, IN DWORD SizeOfDll );
typedef DWORD (__stdcall *SYMSETOPTIONSPROC)(IN DWORD SymOptions);
typedef BOOL (__stdcall *STACKWALKPROC)(DWORD MachineType, HANDLE hProcess,HANDLE hThread, LPSTACKFRAME StackFrame, PVOID ContextRecord,	PREAD_PROCESS_MEMORY_ROUTINE ReadMemoryRoutine,	PFUNCTION_TABLE_ACCESS_ROUTINE FunctionTableAccessRoutine,	PGET_MODULE_BASE_ROUTINE GetModuleBaseRoutine,	PTRANSLATE_ADDRESS_ROUTINE TranslateAddress );
typedef DWORD (__stdcall WINAPI *UNDECORATESYMBOLNAMEPROC)(PCSTR DecoratedName, PSTR UnDecoratedName,DWORD UndecoratedLength, DWORD Flags );
typedef BOOL (__stdcall *SYMENUMERATESYMBOLSPROC)(IN HANDLE hProcess, IN DWORD base,IN PVOID proc,IN PVOID dat);
//
typedef HANDLE (__stdcall *CREATETOOLHELP32SNAPSHOTPROC)( DWORD dwFlags, DWORD th32ProcessID );
typedef BOOL (__stdcall *MODULE32FIRSTPROC)( HANDLE hSnapshot, LPMODULEENTRY32 lpme );
typedef BOOL (__stdcall *MODULE32NEXTPROC)( HANDLE hSnapshot, LPMODULEENTRY32 lpme );
//
typedef struct _MODULEINFO {LPVOID lpBaseOfDll; DWORD SizeOfImage; LPVOID EntryPoint;} MODULEINFO, *LPMODULEINFO;
typedef BOOL (__stdcall *ENUMPROCESSMODULESPROC)( HANDLE hProcess, HMODULE *lphModule, DWORD cb, LPDWORD lpcbNeeded );
typedef DWORD (__stdcall *GETMODULEFILENAMEEXPROC)( HANDLE hProcess, HMODULE hModule, LPSTR lpFilename, DWORD nSize );
typedef DWORD (__stdcall *GETMODULEBASENAMEPROC)( HANDLE hProcess, HMODULE hModule, LPSTR lpFilename, DWORD nSize );
typedef BOOL (__stdcall *GETMODULEINFORMATIONPROC)( HANDLE hProcess, HMODULE hModule, LPMODULEINFO pmi, DWORD nSize );


SYMCLEANUPPROC             pSymCleanup = NULL;
SYMFUNCTIONTABLEACCESSPROC pSymFunctionTableAccess = NULL;
SYMGETLINEFROMADDRPROC     pSymGetLineFromAddr = NULL;
SYMGETMODULEBASEPROC       pSymGetModuleBase = NULL;
SYMGETMODULEINFOPROC       pSymGetModuleInfo = NULL;
SYMGETOPTIONSPROC          pSymGetOptions = NULL;
SYMGETSYMFROMADDRPROC      pSymGetSymFromAddr = NULL;
SYMINITIALIZEPROC          pSymInitialize = NULL;
SYMLOADMODULEPROC          pSymLoadModule = NULL;
SYMSETOPTIONSPROC          pSymSetOptions = NULL;
STACKWALKPROC              pStackWalk = NULL;
UNDECORATESYMBOLNAMEPROC   pUnDecorateSymbolName = NULL;
SYMENUMERATESYMBOLSPROC    pSymEnumerateSymbols = NULL;
//
CREATETOOLHELP32SNAPSHOTPROC pCreateToolhelp32Snapshot = NULL;
MODULE32FIRSTPROC pModule32First = NULL;
MODULE32NEXTPROC pModule32Next = NULL;
//
ENUMPROCESSMODULESPROC pEnumProcessModules = NULL;
GETMODULEFILENAMEEXPROC pGetModuleFileNameEx = NULL;
GETMODULEBASENAMEPROC pGetModuleBaseName = NULL;
GETMODULEINFORMATIONPROC pGetModuleInformation = NULL;






//=============================================================================
// Just some debugging functions
//=============================================================================
//
void __fastcall db(char *c);
char le_text[1000];
char* __fastcall le()
{ FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,NULL,GetLastError(),MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),le_text,999,NULL);
  int i=strlen(le_text)-1;
  while (i>=0 && (le_text[i]=='\r' || le_text[i]=='\n')) {le_text[i]='\0'; i--;}
  return le_text;
}

void __fastcall mpstrcpy(char *dst,char *src)
{ strncpy(dst,src,MAX_PATH); dst[MAX_PATH-1]='\0';
}
void __fastcall mpstrcat(char *dst,char *src)
{ strncat(dst, src, MAX_PATH-strlen(dst)-1);
}

bool __fastcall db_FileExists(const char *c)
{ HANDLE hf=CreateFile(c,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
  if (hf==INVALID_HANDLE_VALUE) return false; CloseHandle(hf); return true;
}

void __fastcall db_init()
{ if (db_isinited) return; db_isinited=true; db_f=NULL;
  bool logtofile = (LogName!=NULL && strlen(LogName)>0);
  if (LogName[0]=='*') logtofile=db_FileExists(&LogName[1]);
  if (logtofile)
  { bool oldval=InternalCall; InternalCall=true;
    if (LogName[0]=='*') db_f=fopen(&LogName[1],"wt");
    else db_f=fopen(LogName,"wt");
    InternalCall=oldval;
  }
  char c[1000];
  if (logtofile) wsprintf(c,"******************* START OF LOG also logged to %s",LogName);
  else wsprintf(c,"******************* START OF LOG");
  db(c);
  struct date d; getdate(&d); struct time t; gettime(&t);
  wsprintf(c,"%s on %i:%i:%i:%i on %i/%i/%i",GetCommandLine(),(int)t.ti_hour,(int)t.ti_min,(int)t.ti_sec,(int)t.ti_hund,(int)d.da_day,(int)d.da_mon,(int)d.da_year);
  db(c);
}
void __fastcall db_exit()
{ if (!db_isinited) return;
  bool oldval=InternalCall; InternalCall=true;
  if (db_f!=NULL) fclose(db_f); db_f=NULL;
  InternalCall=oldval;
  db_isinited=false;
}
void __fastcall db(char *c)
{ db_init();
  OutputDebugString(c);
  if (db_f!=NULL) {fputs(c,db_f); fputs("\n",db_f);}
}
void __fastcall dbcs(TCallstack &cs)
{ if (cs.Depth==0) {db("<no call stack>");return;}
  for (int i=0; i<cs.Depth; i++) db(cs.l[i].c);
}
void __fastcall callstack(TCallstack &cs,CONTEXT &ctx);
AnsiString __fastcall dcallstack()
{ CONTEXT ctx; ZeroMemory(&ctx,sizeof(ctx)); ctx.ContextFlags=CONTEXT_FULL; GetThreadContext(GetCurrentThread(),&ctx);
  TCallstack cs; callstack(cs,ctx); cs.Serious = !DontTakeSeriously;
  AnsiString s=""; if (cs.Depth==0) s="<no call stack>";
  for (int i=0; i<cs.Depth; i++) s=s+AnsiString(cs.l[i].c)+"\r\n";
  return s;
}









//=============================================================================
// SYM SUBSYSTEM
// sym_usemods means either hToolHelp or hPsapi is loaded and will enumerate modlues
// if that's true, then sym_usesyms might be true: to say that hImagehlp is loaded
// If hImagehlp is loaded, then SymInitialized has been closed.
//
// FillModuleList -- gets a list of all modules (i.e. dlls, current exe, packages ...)
//   There are two implementations. First, we try to get it using toolhelp.
//   If that fails, then we try using PSAPI. We load both toolhelp and PSAPI
//   dynamically.
// EnsureModuleSymbolsLoaded - we use 'modcount' as a sort of checksum, recording
//   how many modules' symbols have been loaded. Initially zero. Every time
//   someone calls this function we check how many modules are currently loaded.
//   If there are != modcount, then someone's been loading or unloading dlls,
//   so we take the opportunity to FillModuleList and load symbols from every
//   single one of them again. Obviously the checksum might break, if you
//   unload one library and load another. Be careful!
//=============================================================================
//
int __fastcall sym_FillModuleListTH32(TModuleEntry *mes,DWORD pid)
{ HANDLE hSnap=pCreateToolhelp32Snapshot(TH32CS_SNAPMODULE,pid);
  if (hSnap==(HANDLE)-1) return 0;
  MODULEENTRY32 me; ZeroMemory(&me,sizeof(me)); me.dwSize=sizeof(me);
  int i=0;
  //
  bool keepgoing = pModule32First(hSnap,&me);
  while (keepgoing)
  { mpstrcpy(mes[i].imageName, me.szExePath);
    mpstrcpy(mes[i].moduleName,me.szModule);
    mes[i].baseAddress = (DWORD)me.modBaseAddr;
    mes[i].size = me.modBaseSize;
    //db("snapshot modbase=0x"+IntToHex((int)e->baseAddress,8)+" modbasesize=0x"+IntToHex((int)e->size,8)+" module="+e->moduleName+" exepath="+e->imageName);
    keepgoing = pModule32Next(hSnap,&me);
    i++;
    if (i==MaxModules) keepgoing=false;
  }
  CloseHandle(hSnap);
  return i;
}

int __fastcall sym_FillModuleListPSAPI(TModuleEntry *mes,DWORD pid, HANDLE hProcess )
{ HMODULE hMods[MAX_PATH];
  BOOL bres; DWORD size;
  bres=pEnumProcessModules(hProcess,hMods,MAX_PATH,&size); if (!bres) return 0;
  if (size>MAX_PATH) return 0; //{db("EnumProcessModules failed: not enough space.");
  int nummods = size/sizeof(HMODULE);
  if (nummods>MaxModules) nummods=MaxModules;
  for (int i=0; i<nummods; i++)
  { MODULEINFO mi;
    pGetModuleInformation(hProcess,hMods[i],&mi,sizeof(mi));
    mes[i].baseAddress=(DWORD)mi.lpBaseOfDll;
    mes[i].size=mi.SizeOfImage;
    mes[i].imageName[0]='\0'; pGetModuleFileNameEx(hProcess,hMods[i],mes[i].imageName,MAX_PATH);
    mes[i].moduleName[0]='\0'; pGetModuleBaseName(hProcess,hMods[i],mes[i].moduleName,MAX_PATH);
    //db("snapshot modbase=0x"+IntToHex((int)e->baseAddress,8)+" modbasesize=0x"+IntToHex((int)e->size,8)+" module="+e->moduleName+" exepath="+e->imageName);
  }
  return nummods;
}

void __fastcall sym_EnsureModuleSymbolsLoaded()
{ if (!sym_usemods) return;
  HANDLE hProcess=GetCurrentProcess(); DWORD pid=GetCurrentProcessId();
  TModuleEntry mes[MaxModules]; int newnum=0;
  if (sym_hToolHelp!=NULL) newnum=sym_FillModuleListTH32(mes,pid);
  else if (sym_hPsapi!=NULL) newnum=sym_FillModuleListPSAPI(mes,pid,hProcess);
  //
  // only reload symbols if our naive checksum has indicated a change
  bool haschanged=false;
  if (newnum!=NumMods) haschanged=true;
  if (!haschanged)
  { haschanged = (0 != memcmp(mes,Mods,newnum*sizeof(TModuleEntry)));
  }
  if (!haschanged) return;
  //
  memcpy(Mods,mes,newnum*sizeof(TModuleEntry)); NumMods=newnum;
  for (int i=0; i<NumMods; i++)
  { pSymLoadModule(hProcess,0,Mods[i].imageName,Mods[i].moduleName,Mods[i].baseAddress,Mods[i].size);
    //if (bres) db("Symbols loaded: "+mod->moduleName+" ["+dt+"ms]");
    //else dble("Failed to load symbols for "+mod->moduleName+" ["+dt+"ms]");
  }
}








//=============================================================================
// callstack stuff -- to get a readout of the current call-stack.
//   (involves getting an enumeration of all current processes, either
//   with toolhelp32 or psapi; loading symbol information for each;
//   loading imagehlp; using it to get a stack-trace.
// All the dlls are loaded dynamically, the first time they're needed.
//   Thus, if you never call dcallstack() in your application, there won't
//   be any slowdown.
// There's lots of OutputDebugString that happen. That shouldn't be a problem!
//   After all, if your problem is serious enough for you to need a callstack,
//   then you want all the diagnostics you can have!
//
// IMPORTANT -- to get the callstack it uses standard call-stack routines
//   that are built into windows. These require MS-format debug information,
//   which can be found in .DBG files. Borland does not generate MS-compatible
//   debug files. So, you need to generate a .DBG file yourself. Read the
//   readme.txt for an explanation of how to get them, and a discussion.
//
// sym_init - This function gets called automatically the first time you try
//   to get the callstack. It loads the imagehlp library dynamically,
//   sets up the symbol manager, loads in symbols for all running processes.
// sym_exit - It cleanups the symbol manager and unloads the dll
//=============================================================================
//
void __fastcall sym_init()
{ if (sym_isinited) return; sym_isinited=true; sym_issucc=false; sym_usemods=false; sym_usesyms=false;
  //db("Initing...");
  sym_hImagehlpDll=LoadLibrary("imagehlp.dll"); if (sym_hImagehlpDll==NULL) return; //{db("Failed to load library.");return;}
  pSymCleanup = (SYMCLEANUPPROC) GetProcAddress( sym_hImagehlpDll, "SymCleanup" );
  pSymFunctionTableAccess = (SYMFUNCTIONTABLEACCESSPROC) GetProcAddress( sym_hImagehlpDll, "SymFunctionTableAccess" );
  pSymGetLineFromAddr = (SYMGETLINEFROMADDRPROC) GetProcAddress( sym_hImagehlpDll, "SymGetLineFromAddr" );
  pSymGetModuleBase = (SYMGETMODULEBASEPROC) GetProcAddress( sym_hImagehlpDll, "SymGetModuleBase" );
  pSymGetModuleInfo = (SYMGETMODULEINFOPROC) GetProcAddress( sym_hImagehlpDll, "SymGetModuleInfo" );
  pSymGetOptions = (SYMGETOPTIONSPROC) GetProcAddress( sym_hImagehlpDll, "SymGetOptions" );
  pSymGetSymFromAddr = (SYMGETSYMFROMADDRPROC) GetProcAddress( sym_hImagehlpDll, "SymGetSymFromAddr" );
  pSymInitialize = (SYMINITIALIZEPROC) GetProcAddress( sym_hImagehlpDll, "SymInitialize" );
  pSymSetOptions = (SYMSETOPTIONSPROC) GetProcAddress( sym_hImagehlpDll, "SymSetOptions" );
  pStackWalk = (STACKWALKPROC) GetProcAddress( sym_hImagehlpDll, "StackWalk" );
  pUnDecorateSymbolName = (UNDECORATESYMBOLNAMEPROC) GetProcAddress( sym_hImagehlpDll, "UnDecorateSymbolName" );
  pSymLoadModule = (SYMLOADMODULEPROC) GetProcAddress( sym_hImagehlpDll, "SymLoadModule" );
  pSymEnumerateSymbols = (SYMENUMERATESYMBOLSPROC) GetProcAddress( sym_hImagehlpDll, "SymEnumerateSymbols");
  bool ok=true;
  if (pSymCleanup==NULL) ok=false;
  if (pSymFunctionTableAccess==NULL) ok=false;
  //if (pSymGetLineFromAddr==NULL) ok=false; // This function isn't essential.
  if (pSymGetModuleBase==NULL) ok=false;
  if (pSymGetModuleInfo==NULL) ok=false;
  if (pSymGetOptions==NULL) ok=false;
  if (pSymGetSymFromAddr==NULL) ok=false;
  if (pSymInitialize==NULL) ok=false;
  if (pSymSetOptions==NULL) ok=false;
  if (pStackWalk==NULL) ok=false;
  if (pUnDecorateSymbolName==NULL) ok=false;
  if (pSymLoadModule==NULL) ok=false;
  if (pSymEnumerateSymbols==NULL) ok=false;
  if (!ok) {FreeLibrary(sym_hImagehlpDll); return;} //db("Failed to find all imagehlp functions.");return;}
  //
  // Initialise one of our module-walking friends
  // I think the toolhelp DLL is called tlhelp32.dll on Win9X, so we try both
  //db("Trying to load toolhelp...");
  const char *dllname[] = { "kernel32.dll", "tlhelp32.dll" };
  for (int i=0; i<2; i++)
  { sym_hToolHelp=LoadLibrary(dllname[i]);
    if (sym_hToolHelp==NULL) continue;
    pCreateToolhelp32Snapshot = (CREATETOOLHELP32SNAPSHOTPROC)GetProcAddress(sym_hToolHelp,"CreateToolhelp32Snapshot");
    pModule32First = (MODULE32FIRSTPROC)GetProcAddress(sym_hToolHelp,"Module32First");
    pModule32Next = (MODULE32NEXTPROC)GetProcAddress(sym_hToolHelp,"Module32Next");
    if (pCreateToolhelp32Snapshot!=0 && pModule32First!=0 && pModule32Next!=0) break;
    FreeLibrary(sym_hToolHelp); sym_hToolHelp=NULL;
  }
  if (sym_hToolHelp==NULL)
  { //db("Trying to load PSAPI...");
    sym_hPsapi = LoadLibrary("psapi.dll");
    pEnumProcessModules = (ENUMPROCESSMODULESPROC)GetProcAddress(sym_hPsapi,"EnumProcessModules");
    pGetModuleFileNameEx = (GETMODULEFILENAMEEXPROC)GetProcAddress(sym_hPsapi,"GetModuleFileNameExA");
    pGetModuleBaseName = (GETMODULEBASENAMEPROC)GetProcAddress(sym_hPsapi,"GetModuleBaseNameA");
    pGetModuleInformation = (GETMODULEINFORMATIONPROC)GetProcAddress(sym_hPsapi,"GetModuleInformation");
    if (pEnumProcessModules==0 || pGetModuleFileNameEx==0 || pGetModuleBaseName==0 || pGetModuleInformation==0)
    { FreeLibrary(sym_hPsapi); sym_hPsapi=NULL;
    }
  }
  sym_usemods = (sym_hToolHelp!=NULL || sym_hPsapi!=NULL);
  if (!sym_usemods)
  { //db("Could not load either toolhelp or psapi. There will be no symbol information.");
  }
  //
  // build symbol search: name of apps directory, without trailing slash, plus env.vars
  if (sym_usemods)
  { char sympath[MAX_PATH*8], buf[MAX_PATH]; DWORD dres;
    GetModuleFileName(NULL,sympath,MAX_PATH);
    int i=strlen(sympath)-1; while (i>0 && sympath[i]!='\\') i--; if (i!=0) sympath[i]=0;
    dres=GetEnvironmentVariable("_NT_SYMBOL_PATH", buf,MAX_PATH); if (dres>0) {strcat(sympath,";");strcat(sympath,buf);}
    dres=GetEnvironmentVariable("_NT_ALTERNATE_SYMBOL_PATH",buf,MAX_PATH); if (dres>0) {strcat(sympath,";");strcat(sympath,buf);}
    dres=GetEnvironmentVariable("SYSTEMROOT",buf,MAX_PATH); if (dres>0)
    { strcat(sympath,";");
      strcat(sympath,buf);
    }
    //db("sympath: "+sympath);
    sym_usesyms = pSymInitialize(GetCurrentProcess(),sympath,false);
    if (sym_usesyms)
    { DWORD symOptions = pSymGetOptions();
      symOptions |= SYMOPT_LOAD_LINES; symOptions &= ~SYMOPT_UNDNAME;
      pSymSetOptions(symOptions);
    }
    else
    { FreeLibrary(sym_hImagehlpDll); sym_hImagehlpDll=NULL;
      if (sym_hToolHelp!=NULL) FreeLibrary(sym_hToolHelp); sym_hToolHelp=NULL;
      if (sym_hPsapi!=NULL) FreeLibrary(sym_hPsapi); sym_hPsapi=NULL;
      //db("Syms failed to initialize. There will be no symbol information");
    }
  }
  //
  //
  sym_issucc=true;
}

void __fastcall sym_exit()
{ if (!sym_isinited) return;
  //db("Terminating...");
  if (sym_hToolHelp!=NULL) FreeLibrary(sym_hToolHelp); sym_hToolHelp=NULL;
  if (sym_hPsapi!=NULL) FreeLibrary(sym_hPsapi); sym_hPsapi=NULL;
  if (sym_hImagehlpDll!=NULL)
  { pSymCleanup(GetCurrentProcess());
    FreeLibrary(sym_hImagehlpDll); sym_hImagehlpDll=NULL;
  }
  sym_isinited=false;
}













//=============================================================================
// cs_callstack - retrieves the current call stack. Uses symbol information
//   if present as well.
//=============================================================================
//
void __fastcall callstack(TCallstack &cs,CONTEXT &ctx)
{ cs.Depth=0; sym_init(); if (!sym_issucc) return; //"<failed to init debugging>";
  HANDLE hProcess=GetCurrentProcess(), hThread=GetCurrentThread();
  sym_EnsureModuleSymbolsLoaded();
  //
  STACKFRAME s; ZeroMemory(&s,sizeof(s));
  s.AddrPC.Offset = ctx.Eip;    s.AddrPC.Mode = AddrModeFlat;
  s.AddrFrame.Offset = ctx.Ebp; s.AddrFrame.Mode = AddrModeFlat;
  char symbuf[sizeof(IMAGEHLP_SYMBOL) + MAX_PATH]; IMAGEHLP_SYMBOL *pSym=(IMAGEHLP_SYMBOL*)symbuf;
  ZeroMemory(pSym,sizeof(IMAGEHLP_SYMBOL)+MAX_PATH); pSym->SizeOfStruct=sizeof(IMAGEHLP_SYMBOL); pSym->MaxNameLength=MAX_PATH;
  IMAGEHLP_LINE Line; ZeroMemory(&Line,sizeof(Line)); Line.SizeOfStruct=sizeof(Line);
  IMAGEHLP_MODULE Module; ZeroMemory(&Module,sizeof(Module)); Module.SizeOfStruct=sizeof(Module);
  //
  while (cs.Depth<MaxStackDepth)
  { int nframe=cs.Depth;
    //
    bool bres=pStackWalk(IMAGE_FILE_MACHINE_I386,hProcess,hThread,&s,&ctx,NULL,pSymFunctionTableAccess,pSymGetModuleBase,NULL);
    if (!bres) break;
    cs.l[nframe].pc = (void*)s.AddrPC.Offset;
    cs.l[nframe].frame = (void*)s.AddrFrame.Offset;
    wsprintf(cs.l[nframe].c, "%lx ",s.AddrPC.Offset);
    //
    if (s.AddrPC.Offset!=0 && sym_usesyms)
    { // First, the symbol
      ZeroMemory(pSym,sizeof(IMAGEHLP_SYMBOL)+MAX_PATH); pSym->SizeOfStruct=sizeof(IMAGEHLP_SYMBOL); pSym->MaxNameLength=MAX_PATH;
      DWORD offsetFromSymbol=0;
      bres=pSymGetSymFromAddr(hProcess, s.AddrPC.Offset,&offsetFromSymbol,pSym);
      if (bres)
      { char undName[MAX_PATH], undFullName[MAX_PATH];
        pUnDecorateSymbolName(pSym->Name,undName,MAX_PATH,UNDNAME_NAME_ONLY);
        pUnDecorateSymbolName(pSym->Name,undFullName,MAX_PATH,UNDNAME_COMPLETE);
        mpstrcat(cs.l[nframe].c, undFullName);
      }
      // Third, the line number.
      bres=false;
      if (pSymGetLineFromAddr!=NULL) bres=pSymGetLineFromAddr(hProcess,s.AddrPC.Offset,&offsetFromSymbol,&Line);
      if (bres)
      { char buf[50]; wsprintf(buf," (%s +%i)",Line.FileName,(int)Line.LineNumber);
        mpstrcat(cs.l[nframe].c, buf);
      }
      // Fourth, the module info
      bres=SymGetModuleInfo(hProcess,s.AddrPC.Offset,&Module);
      if (bres)
      { char buf[50]; wsprintf(buf," [%s]",Module.ModuleName);
        mpstrcat(cs.l[nframe].c, buf);
      }
    } // end if (usesyms)
    if (s.AddrReturn.Offset==0) {SetLastError(0); break;}
    //
    cs.Depth++;
  }
  //
  //if (GetLastError()!=0) dble("StackWalk: had some error or other.");
}





















void MyExitProc();
void MyStartProc();
void MyStartTakingSeriouslyProc();
#pragma exit MyExitProc 0
#pragma startup MyStartProc 0
#pragma startup MyStartTakingSeriouslyProc 32


void* __fastcall MyGetMem(int i)
{ BYTE *real=new BYTE[4+i];
  *((int*)real)=i;
  return real+4;
}
int __fastcall MyFreeMem(void *p)
{ BYTE *fake=(BYTE*)p;
  BYTE *real=fake-4;
  delete[] real;
  return 0;
}
void* __fastcall MyReallocMem(void *p,int s)
{ BYTE *oldfake=(BYTE*)p;
  BYTE *oldreal=oldfake-4;
  int oldsize=*((int*)oldreal);
  BYTE *newreal=new BYTE[4+s];
  *((int*)newreal)=s;
  BYTE *newfake=newreal+4;
  int max=oldsize; if (max>s) max=s;
  memcpy(newfake,oldfake,max);
  delete[] oldreal;
  return newreal+4;
}
TMemoryManager MyMemoryManager = {MyGetMem,MyFreeMem,MyReallocMem};
void MyStartTakingSeriouslyProc() {DontTakeSeriously=false;}
void MyStartProc()
{
  #pragma option push -w-ccc -w-rch
  if (WatchVcl) {SetMemoryManager(MyMemoryManager);}
  #pragma option pop

}







void * operator new (size_t Size)
{ if (InternalCall) return malloc(Size);
  //
  CONTEXT ctx; ZeroMemory(&ctx,sizeof(ctx)); ctx.ContextFlags=CONTEXT_FULL; GetThreadContext(GetCurrentThread(),&ctx);
  TCallstack cs; callstack(cs,ctx); cs.Serious = !DontTakeSeriously;
  char *RealMem = (char*)malloc(LoSafetySize + Size + HiSafetySize + sizeof(TCallstack));
  if (RealMem==NULL)
  { db("********* ERROR ****************");
    db("Out of memory");
    return RealMem;
  }
  //
  char *FakeMem = RealMem + LoSafetySize;
  if (MemsFull) return FakeMem;
  //
  FillMemory(RealMem,LoSafetySize,SafetyChar);
  FillMemory((LPBYTE)FakeMem+Size,HiSafetySize,SafetyChar);
  memcpy(RealMem + LoSafetySize+Size+HiSafetySize, &cs, sizeof(TCallstack));
  //
  // Find somewhere for it in our database
  int i=0; while (i<NumEnts && Mems[i].Address!=NULL) i++;
  if (i==NumEnts)
  { MemsFull=true;
    db("******** SERIOUS ERROR ***************");
    db("Memory database too small - increase NumEnts");
    db("All following messages may be incorrect");
    return FakeMem;
  }
  //
  if (i>MaxUsed) MaxUsed=i;
  if (!DontTakeSeriously) NumNew++;
  Mems[i].Address=FakeMem;
  Mems[i].Size=Size;
  return FakeMem;
}
void * operator new[] (size_t Size) {return operator new(Size);}


void operator delete (void *vFakeMem)
{ if (InternalCall) {if (vFakeMem) free(vFakeMem);return;}
  //int res=heapcheck();
  //if (res==_HEAPCORRUPT)
  //{ db("*************** ERROR ************************");
  //  db("The runtime library reported a corrupt heap");
  //  db("All bets are off");
  //}
  if (vFakeMem==NULL) NumNullDeletes++;
  else  NumDeletes++;
  if (vFakeMem==NULL && ReportNulls)
  { db("Warning... NULL pointer deleted... this is valid but suspicious");
    CONTEXT ctx; ZeroMemory(&ctx,sizeof(ctx)); ctx.ContextFlags=CONTEXT_FULL; GetThreadContext(GetCurrentThread,&ctx);
    TCallstack cs; callstack(cs,ctx);
    db("It was deleted at");
    dbcs(cs);
  }
  if (vFakeMem==NULL) return;
  //
  // Check that pointer is in our database
  int i=0; while (i<NumEnts && Mems[i].Address!=vFakeMem) i++;
  if (i==NumEnts)
  { char c[100]; wsprintf(c,"Unknown pointer 0x%lx deleted",vFakeMem);
    db("******************* ERROR *******************");
    db(c);
    db("Either the pointer is corrupt or was deleted");
    db("multiple times.");
    CONTEXT ctx; ZeroMemory(&ctx,sizeof(ctx)); ctx.ContextFlags=CONTEXT_FULL; GetThreadContext(GetCurrentThread(),&ctx);
    TCallstack cs; callstack(cs,ctx);
    db("It was deleted at");
    dbcs(cs);
    #pragma option push -w-ccc
    if (NagScreen)
    { char c[200]; wsprintf(c,"An invalid pointer is being deleted. Pause now to look at the callstack.");
      if (db_f!=NULL) {strcat(c,"\r\nA log is being written to ");strcat(c,LogName);}
      MessageBox(NULL,c,"*VERY* serious error",MB_OK);
    }
    #pragma option pop
    free(vFakeMem);
    return;
  }
  //
  char *RealMem=(char*)vFakeMem - LoSafetySize;
  char *locheckmem=RealMem;
  char *FakeMem=RealMem+LoSafetySize;
  char *hicheckmem=RealMem+LoSafetySize+Mems[i].Size;
  TCallstack &cs=*(TCallstack*)(RealMem+LoSafetySize+Mems[i].Size+HiSafetySize);
  // If it was in our database, and it wasn't serious, then we won't have recorded
  // it as a numnew in the first place. Just so the things tally, we will now
  if (!cs.Serious) NumNew++;
  // Check safety areas
  bool UnderWritten=false, OverWritten=false;
  for (int i=0; i<LoSafetySize; i++) if (locheckmem[i]!=SafetyChar) UnderWritten=true;
  for (int i=0; i<HiSafetySize; i++) if (hicheckmem[i]!=SafetyChar) OverWritten=true;
  if (OverWritten || UnderWritten)
  { char c[100];
    wsprintf(c,"The memory block at 0x%lx is corrupt",FakeMem);
    db("****************** WARNING ******************");
    db(c);
    if (UnderWritten) db("The low safety area has been changed");
    if (OverWritten) db("The high safety area has been changed.");
    db("It was created at");
    dbcs(cs);
    CONTEXT ctx; ZeroMemory(&ctx,sizeof(ctx)); ctx.ContextFlags=CONTEXT_FULL; GetThreadContext(GetCurrentThread(),&ctx);
    TCallstack cs; callstack(cs,ctx);
    db("It was deleted at");
    dbcs(cs);
    char *msg="";
    if (OverWritten && !UnderWritten) msg="Block had overrun";
    if (!OverWritten && UnderWritten) msg="Block had been underwritten";
    if (OverWritten && UnderWritten) msg="Block had overrun below and above";
    #pragma option push -w-ccc
    if (NagScreen) MessageBox(NULL,msg,"This is terrible",MB_OK);
    #pragma option pop
  }
  //
  Mems[i].Address = NULL;
  free(RealMem);
}
void operator delete[] (void * TheMem) {delete TheMem;}



void MyExitProc()
{ bool oldval=InternalCall;
  InternalCall=true;
  int leakage=0;
  //
  // Everything left in the database is a memory leak!
  for (int i=0;i<MaxUsed;i++)
  { char *FakeMem=(char*)Mems[i].Address;
    char *RealMem=FakeMem-LoSafetySize;
    TCallstack &cs = *(TCallstack*)(RealMem+LoSafetySize+Mems[i].Size+HiSafetySize);
    if (FakeMem!=NULL && cs.Serious)
    { char c[100]; wsprintf(c,"Memory block at 0x%lx not deleted",Mems[i].Address);
      char d[100]; wsprintf(d,"It is %i bytes long",Mems[i].Size);
      db("******************* ERROR *******************");
      db(c); db(d);
      db("It was created at");
      dbcs(cs);
      leakage=leakage+Mems[i].Size;
    }
  }
  /*
  int res=_heapchk();
  if (res==_HEAPCORRUPT)
  { db("**************** ERROR *******************");
    db("run-time library reports that the heap is corrupt");
    db("All bets are off");
  }
  if (res==_HEAPOK)
  { db("************* ERROR ********************");
    db("run-time library reports that the heap is not empty");
    _HEAPINFO hi;
    hi._pentry=NULL;
    int res=_rtl_heapwalk(&hi);
    while (res==_HEAPOK)
    { char c[100]; wsprintf(c,"rtl: heap at 0x%lx, size %i",hi._pentry,hi._useflag);
      db(c);
      res=_rtl_heapwalk(&hi);
    }
  }
  */
  char c[100]; wsprintf(c,"Max database entries used: %i",MaxUsed+1);
  char d[100]; wsprintf(d,"Calls to new: %i",NumNew);
  char e[100]; wsprintf(e,"Calls to delete (non-null): %i",NumDeletes);
  char f[100]; wsprintf(f,"Calls to delete (null): %i",NumNullDeletes);
  char g[100]; wsprintf(g,"Memory leakage: %i",leakage);
  char h[100]; strcpy(h,g); if (db_f!=NULL) {strcat(h,"\r\nLog written to ");strcat(h,LogName);}
  db("**************** INFORMATION ****************");
  #pragma option push -w-ccc
  db(c); db(d); db(e); if (ReportNulls) db(f); db(g);
  #pragma option pop
  if (leakage>0 && NagScreen) MessageBox(NULL,h,"This isn't too good...",MB_OK);
  //
  sym_exit();
  db_exit();
  //
  InternalCall=oldval;
}








bool __fastcall CheckSafetyAreas(void *FakeMem)
{ if (FakeMem==NULL) return true;
  void *RealMem=(char*)FakeMem - LoSafetySize;
  //int res=heapcheck();
  //if (res==_HEAPCORRUPT)
  //{ db("*************** ERROR ************************");
  //  db("The runtime library reported a corrupt heap");
  //  db("All bets are off");
  //}
  //
  // Check that pointer is in our database
  int i; for (i=0;i<NumEnts && Mems[i].Address!=FakeMem;i++);
  if (i==NumEnts)
  { char c[100]; wsprintf(c,"CheckMem: Pointer 0x%lx couldn't be found",FakeMem);
    if (!MemsFull) db("******************* ERROR *******************");
    db(c);
    db("It's therefore not possible to check if its safety areas were corrupted");
    return false;
  }
  // Check safety areas
  LPBYTE locheckmem=(LPBYTE)RealMem;
  LPBYTE hicheckmem=(LPBYTE)FakeMem+Mems[i].Size;
  TCallstack &cs=*(TCallstack*)((LPBYTE)hicheckmem+HiSafetySize);
  bool UnderWritten=false, OverWritten=false;
  for (int i=0; i<LoSafetySize; i++) if (locheckmem[i]!=SafetyChar) UnderWritten=true;
  for (int i=0; i<HiSafetySize; i++) if (hicheckmem[i]!=SafetyChar) OverWritten=true;
  if (!OverWritten && !UnderWritten) return true;
  //
  char c[100];
  wsprintf(c,"The memory block at 0x%lx is corrupt",FakeMem);
  db("****************** WARNING ******************");
  db(c);
  if (UnderWritten) db("The low safety area has been changed");
  if (OverWritten) db("The high safety area has been changed.");
  db("The memory block in question was created at");
  dbcs(cs);
  char *msg="";
  if (OverWritten && !UnderWritten) msg="Block had overrun";
  if (!OverWritten && UnderWritten) msg="Block had been underwritten";
  if (OverWritten && UnderWritten) msg="Block had overrun below and above";
  #pragma option push -w-ccc
  if (NagScreen) MessageBox(NULL,msg,"This is terrible",MB_OK);
  #pragma option pop
  return false;
}





